/**********************************************************************
Code associated to the paper:

Resolution of a large number of small random symmetric 
linear systems in single precision arithmetic on GPUs

by: 

Lokman A. Abbas-Turki and Stef Graillat

Those who re-use this code should mention in their code 
the name of the authors above.
**********************************************************************/
/******************************************************************
The main kernel DiagSymdiag_k performs the diagonalization
of tridiagonal symmetric matrices that:

1) Do not contain null elements on the subdiagonal.
2) Do not contain elements bigger than 10 in absolute value. It 
is even better that the biggest values are close to 10.

With a simple pre-processing (Complexity=O(n)), one can reduce any 
matrix into matrices (or only one matrix) that satisfy 1) and 2). 
******************************************************************/

#include <stdio.h>
#include <math.h>
#include <stdlib.h>

#define IT 10
#define thr (0.01f)
#define thr1 (0.00001f)
#define thr2 (1.8f*thr1)


// Diagonalization of 2x2 matrices.
__device__ void diagSym2D_d(float *T, float *D, int n){

	
	float Delta, tmp, tmp0, tmp1, tmp2, tmp3;


	Delta = (T[0]-T[n+1])*(T[0]-T[n+1]) + 4.0f*T[n]*T[n];

	D[0] = (T[0]+T[n+1]-sqrtf(Delta))/2.0f;

	D[1] = (T[0]+T[n+1]+sqrtf(Delta))/2.0f;

	tmp = 4.0f*T[n]*T[n];

	tmp0 = sqrtf(tmp/(tmp+(T[0]-T[n+1]+sqrtf(Delta))*(T[0]-T[n+1]+sqrtf(Delta))));
	
	tmp1 = -sqrtf(tmp/(tmp+(T[0]-T[n+1]-sqrtf(Delta))*(T[0]-T[n+1]-sqrtf(Delta))));

	tmp2 = -(T[0]-T[n+1]+sqrtf(Delta))*tmp0/(2.0f*T[1]);

	tmp3 = -(T[0]-T[n+1]-sqrtf(Delta))*tmp1/(2.0f*T[1]);

	T[0] = tmp0;
	T[1] = tmp1;
	T[n] = tmp2;
	T[n+1] = tmp3;
}


// A preprocessing done when previous eigenvalues are close.
__device__ void jacoRot_d(float *y, float *u, float *G, float rho, int i, 
						  int j, int m, int M, int n, float g, 
						  float h){

	int k, I, J;
	float alpha, s, c, tmp1, tmp2;

	if(rho>0){
		J = j;
		I = i;
	}else{
		J = i;
		I = j;
	}
	if(fabsf(y[i]-y[j])<=4.0f*h*thr2*fmaxf(10.0f,fabsf(y[i]))){
		

		y[J] = y[I];
		alpha = sqrtf(u[I]*u[I]+u[J]*u[J]);
		c = alpha/(u[I]*(1.0f+u[J]*u[J]/(u[I]*u[I])));
		s = u[J]/alpha;
	   
		// EIG vector rotation
		tmp1 = c*u[J] - s*u[I];
		tmp2 = s*u[J] + c*u[I];
		u[J] = tmp1;
		u[I] = tmp2;

		// matrix rotation 
		for(k=m; k<M; k++){
   			tmp1 = c*G[k*n+J] - s*G[k*n+I];
			tmp2 = s*G[k*n+J] + c*G[k*n+I];
			G[k*n+J] = tmp1;
			G[k*n+I] = tmp2;
		}

	}
}


// Sorting previous eigenvalue.
__device__ void eigSort_d(float *x, float *U, int *b, int i){

	float tmp;
	int tmpI;

	if(x[i]>x[i+1]+thr1/100.0f){
		tmp = x[i];
        x[i] = x[i+1];
        x[i+1] = tmp;
        tmp = U[i];
		U[i] = U[i+1];
        U[i+1] = tmp;
        tmpI = b[i];
		b[i] = b[i+1];
        b[i+1] = tmpI;
	}
}


// Using Lowner's transformation.
__device__ void lowner_d(float *a, float *x, float *U, float *u, 
						 float *y, float rho, int i, int n, 
						 float g, float h){
	
	int j;

	if(rho>0.0f){
		if(fabsf(U[i])<g*thr1){
			u[i]=0.0f;
		}else{
			u[i]=1.0f;
			for(j=0; j<n; j++){
				if(fabsf(x[i]-a[j])>h*thr2){
				   u[i] = u[i]*(a[j]-x[i]);
				}
				if(fabsf(x[i]-y[j])>h*thr2){
				   u[i] = u[i]/(y[j]-x[i]);
				}
			}
			u[i] = copysignf(sqrtf(fabsf(u[i])), U[i]);
		}
	}else{
		if(fabsf(U[i])<g*thr1){
			u[i]=0.0f;
		}else{
			u[i]=1.0f;
			for(j=0; j<n; j++){
				if(fabsf(x[i]-a[j])>h*thr2){
				   u[i] = u[i]*(a[j]-x[i]);
				}
				if(fabsf(x[i]-y[j])>h*thr2){
				   u[i] = u[i]/(y[j]-x[i]);
				}
			}
			u[i] = copysignf(sqrtf(fabsf(u[i])), U[i]);
		}
	}
}


// Set to 0 the components to deflate.
__device__ void deflation_d(float *a, float *x, float *u, float rho, 
					int *b, int i, int n, float h){

	if(rho > 0.0f){
		if(i<n-1){
			if(a[i]-x[i]<=h*thr2){
				u[b[i]] = 0.0f;
			}else{
				if(x[i+1]-a[i]<=h*thr2){
					u[b[i+1]] = 0.0f;
				}
			}	
		}else{
			if(a[i]-x[i]<=h*thr2){
				u[b[i]] = 0.0f;
			}		
		}
	}else{
		if(i>0){
			if(x[i]-a[i]<=h*thr2){
				u[b[i]] = 0.0f;
			}else{
				if(a[i]-x[i-1]<=h*thr2){
					u[b[i-1]] = 0.0f;
				}
			}		
		}else{
			if(x[i]-a[i]<=h*thr2){
				u[b[i]] = 0.0f;
			}		
		}	
	}
}


// Eigenvectors computation.
__device__ void eigVect_d(float *a, float *x, float *u, float *r, 
				  float rho, int *b, int i, int n, int N, float h){

	int j;
	float tmp;
	float sum2;

	if(rho > 0){
		if(i<n-1){
			if(x[b[i+1]]-a[i]>h*thr2 && a[i]-x[b[i]]>h*thr2){
				sum2 = 0.0f;
				for(j=0; j<n; j++){
					tmp = u[j]/(x[j]-a[i]);
					sum2 += tmp*tmp;
					r[i*N+j] = tmp;
				}
				for(j=0; j<n; j++){
					r[i*N+j] /= sqrtf(sum2);
				}
			}else{
				if(a[i]-x[b[i]]<=h*thr2){
					for(j=0; j<n; j++){
						if(j==b[i]){
							r[i*N+j] = 1.0f;
						}else{
							r[i*N+j] = 0.0f;
						}
					}
				}else{
					for(j=0; j<n; j++){
						if(j==b[i+1]){
							r[i*N+j] = 1.0f;
						}else{
							r[i*N+j] = 0.0f;
						}
					}
				}
			}
		}else{
			if(a[i]-x[b[i]]>h*thr2){
				sum2 = 0.0f;
				for(j=0; j<n; j++){
					tmp = u[j]/(x[j]-a[i]);
					sum2 += tmp*tmp;
					r[i*N+j] = tmp;
				}
				for(j=0; j<n; j++){
					r[i*N+j] /= sqrtf(sum2);
				}
			}else{
				for(j=0; j<n; j++){
					if(j==b[i]){
						r[i*N+j] = 1.0f;
					}else{
						r[i*N+j] = 0.0f;
					}
				}			
			}
		}
	}else{
		if(i>0){
			if(x[b[i]]-a[i]>h*thr2 && a[i]-x[b[i-1]]>h*thr2){
				sum2 = 0.0f;
				for(j=0; j<n; j++){
					tmp = u[j]/(x[j]-a[i]);
					sum2 += tmp*tmp;
					r[i*N+j] = tmp;
				}
				for(j=0; j<n; j++){
					r[i*N+j] /= sqrtf(sum2);
				}
			}else{
				if(x[b[i]]-a[i]<=h*thr2){
					for(j=0; j<n; j++){
						if(j==b[i]){
							r[i*N+j] = 1.0f;
						}else{
							r[i*N+j] = 0.0f;
						}
					}
				}else{
					for(j=0; j<n; j++){
						if(j==b[i-1]){
							r[i*N+j] = 1.0f;
						}else{
							r[i*N+j] = 0.0f;
						}
					}
				}
			}
		}else{
			if(x[b[i]]-a[i]>h*thr2){
				sum2 = 0.0f;
				for(j=0; j<n; j++){
					tmp = u[j]/(x[j]-a[i]);
					sum2 += tmp*tmp;
					r[i*N+j] = tmp;
				}
				for(j=0; j<n; j++){
					r[i*N+j] /= sqrtf(sum2);
				}
			}else{
				for(j=0; j<n; j++){
					if(j==b[i]){
						r[i*N+j] = 1.0f;
					}else{
						r[i*N+j] = 0.0f;
					}
				}			
			}
		}	
	}
}


// Transposition of a matrix
__device__  void transLC_d(float *In, float *Out, int j, int n, int N){

	int k;

	for(k=0; k<n; k++){
		Out[j*N+k] = In[k*N+j];
	}
}


// Line to line matrix multiplication: In1*In2'
__device__  void prodLL_d(float *In1, float *In2, int i, int j, 
						  float *sum, int n, int N){

	int k;
	*sum = 0.0f;

	for(k=0; k<n; k++){
		*sum += In1[i*N+k]*In2[j*N+k];
	}
}


// Gragg's algorithm for eigenvalues + Dichotomy for unbounded eigenvalues.
__device__ void gragg_d(float *D, float *U, float rho, float *lam, 
					    int k, int n, int flag, float *fmid){
	
    int i, K;
    float a1, a2, b1, b2, b3, b4, c1, c2, dk, dkp1, di, Disc;

	if(flag==0){
		if(k<n-1 && k>-1){
			a1 = 0.0f;
			a2 = 0.0f;
			dk = (D[k] + D[k+1])/2.0f;
			for(i=0; i<k; i++){
				di = D[i] - dk;
				a1 = a1 + U[i]*U[i]/di;
			}
			for(i=n-1; i>k+1; i--){
				di = D[i] - dk;
				a2 = a2 + U[i]*U[i]/di;
			}
			c1 = a1 + a2 + 1.0f/rho;
			b1 = c1 + U[k]*U[k]/(D[k]-dk);
			b1 = b1 + U[k+1]*U[k+1]/(D[k+1]-dk);
			*fmid = b1;
			dk = D[k+1] - D[k];
			if(b1<0){
				a1 = U[k]*U[k]+U[k+1]*U[k+1]-c1*dk;
				a2 = -U[k+1]*U[k+1]*dk;
				if(a1>0){
					Disc = fmaxf(0.0f, a1*a1-4.0f*a2*c1);
					*lam = D[k+1] + (2.0f*a2)/(a1 + sqrtf(Disc));			
				}else{
					Disc = fmaxf(0.0f, a1*a1-4.0f*a2*c1);
					*lam = D[k+1] + (a1 - sqrtf(Disc))/(2.0f*c1);			
				}
			}else{
				a1 = U[k]*U[k]+U[k+1]*U[k+1]+c1*dk;
				a2 = U[k]*U[k]*dk;
				if(a1>0){
					Disc = fmaxf(0.0f, a1*a1-4.0f*a2*c1);
					*lam = D[k] + (2.0f*a2)/(a1 + sqrtf(Disc));			
				}else{
					Disc = fmaxf(0.0f, a1*a1-4.0f*a2*c1);
					*lam = D[k] + (a1 - sqrtf(Disc))/(2.0f*c1);			
				}		
			}
				}else{
			if(k==n-1){
				b1 = 1.0f/rho;
				b2 = 1.0f/rho;
				dk = D[k]+256.0f;
				for (i=0; i<n; i++){
					di = D[i]-dk;
					b1 = b1 + U[i]*U[i]/di;
				}
				dk = D[k]+128.0f;
				for (i=0; i<n; i++){
					di = D[i]-dk;
					b2 = b2 + U[i]*U[i]/di;
				}
				Disc = 64.0f;
				while(Disc>1){
					if(b1*b2<0){
						b2 = 1.0f/rho;
						dk = dk+Disc;
						for (i=0; i<n; i++){
							di = D[i]-dk;
							b2 = b2 + U[i]*U[i]/di;
						}
					}else{
						b1 = b2;
						b2 = 1.0f/rho;
						dk = dk-Disc;
						for (i=0; i<n; i++){
							di = D[i]-dk;
							b2 = b2 + U[i]*U[i]/di;
						}            
					}
					Disc = Disc/2;
				}
				if(b1*b2<0){
					*lam = dk + Disc;
				}else{*lam = dk - Disc;}			
			}else{
				b1 = 1.0f/rho;
				b2 = 1.0f/rho;
				dk = D[k+1]-256.0f;
				for (i=0; i<n; i++){
					di = D[i]-dk;
					b1 = b1 + U[i]*U[i]/di;
				}
				dk = D[k+1]-128.0f;
				for (i=0; i<n; i++){
					di = D[i]-dk;
					b2 = b2 + U[i]*U[i]/di;
				}
				Disc = 64.0f;
				while(Disc>1){
					if(b1*b2<0){
						b2 = 1.0f/rho;
						dk = dk-Disc;
						for (i=0; i<n; i++){
							di = D[i]-dk;
							b2 = b2 + U[i]*U[i]/di;
						}
					}else{
						b1 = b2;
						b2 = 1.0f/rho;
						dk = dk+Disc;
						for (i=0; i<n; i++){
							di = D[i]-dk;
							b2 = b2 + U[i]*U[i]/di;
						}            
					}
					Disc = Disc/2;
				}
				if(b1*b2<0){
					*lam = dk - Disc;
				}else{*lam = dk + Disc;}			
			}	
		}
	}else{
		if(k==-1){
			K = 0;
		}
		if(k==n-1){
			K = n-2;
		}
		if(k<n-1 && k>-1){
			K = k;
		}
		
		a1 = 0.0f;
		a2 = 0.0f;
		b1 = 0.0f;
		b3 = 0.0f;
		c1 = 0.0f;
		c2 = 0.0f;
		dk = D[K] - *lam;
		dkp1 = D[K+1] - *lam;

		for(i=0; i<=k; i++){
			di = D[i] - *lam;
			if(fabsf(di)>thr*U[i]*U[i]){
				a1 = a1 + U[i]*U[i]/di;
				b1 = b1 + U[i]*U[i]/(di*di);
				c1 = c1 + U[i]*U[i]/(di*di*di);
			}	
		}
		for(i=n-1; i>k; i--){
			di = D[i] - *lam;
			if(fabsf(di)>thr*U[i]*U[i]){
				a2 = a2 + U[i]*U[i]/di;
				b3 = b3 + U[i]*U[i]/(di*di);
				c2 = c2 + U[i]*U[i]/(di*di*di);
			}	
		}
		b2 = dkp1*b1;
		b1 = dk*b1;
		c1 = dk*dkp1*c1;
		b4 = dkp1*b3;
		b3 = dk*b3;
		c2 = dk*dkp1*c2;
		for(i=0; i<=k; i++){
			di = D[i] - *lam;
			if(fabsf(di)<=thr*U[i]*U[i]){
				a1 = a1 + U[i]*U[i]/di;
				b1 = b1 + (dk/di)*U[i]*U[i]/(di);
				b2 = b2 + (dkp1/di)*U[i]*U[i]/(di);
				c1 = c1 + ((dk/di)*(dkp1/di))*U[i]*U[i]/(di);
			}	
		}
		for(i=n-1; i>k; i--){
			di = D[i] - *lam;
			if(fabsf(di)<=thr*U[i]*U[i]){
				a2 = a2 + U[i]*U[i]/di;
				b3 = b3 + (dk/di)*U[i]*U[i]/(di);
				b4 = b4 + (dkp1/di)*U[i]*U[i]/(di);
				c2 = c2 + ((dk/di)*(dkp1/di))*U[i]*U[i]/(di);
			}
		}
		a2 = a1+a2+1.0f/rho;
		b3 = b1+b3;
		b4 = b2+b4;
		c2 = c2+c1;

		b2 = dkp1*dkp1*(b4-c2)/(D[K+1] - D[K]);
		b1 = dk*dk*(c2-b3)/(D[K+1] - D[K]);
		c1 = a2-(b3+b4)+c2;

		a1 = a2*(dk+dkp1)-dkp1*b3;
		a2 = a2*dk*dkp1;

		if (K==k){
			if (a1>0){
				Disc = fmaxf(0.0f, a1*a1-4.0f*a2*c1);
				*lam = *lam + (2.0f*a2)/(a1+sqrtf(Disc));
			}else{
				Disc = fmaxf(0.0f, a1*a1-4.0f*a2*c1);
				*lam = *lam + (a1-sqrtf(Disc))/(2.0f*c1);
			}
		}else{
			if (a1>0){
				Disc = fmaxf(0.0f, a1*a1-4.0f*a2*c1);
				*lam = *lam + (a1+sqrtf(Disc))/(2.0f*c1);
			}else{
				Disc = fmaxf(0.0f, a1*a1-4.0f*a2*c1);
				*lam = *lam + (2.0f*a2)/(a1-sqrtf(Disc));
			}     
		}	
	}
}


// Iterative eigenvalues computation.
__device__ void eigVal_d(float *D, float *U, float *lam, float *u,
						 int *b, float rho, int i, int n, float g){

	float dk, dkp1, lamIn, lamOut, tmp1, tmp2, tmp3, fmid;
	int it = 0;
	

	if((i>n-2) || (i<0)){
		if(i>n-2){
			lamIn = D[i] + rho*rho*0.1f;
			tmp2 = lamIn;
			gragg_d(D, U, rho, &tmp2, n-1, n, 0, &fmid);
			lamOut = fmaxf(tmp2, D[i]+thr1);
			dk = lamOut-D[i];
			tmp1 = lamOut-lamIn;
			while((fabsf(tmp1)>thr1*dk || it<1) && (dk>thr1) && (it<IT)){
				lamIn = lamOut;
				tmp2 = lamIn;
				gragg_d(D, U, rho, &tmp2, n-1, n, 1, &tmp3);
				lamOut = fmaxf(tmp2, D[i]+thr1);
				dk = fminf(dk,lamOut-D[i]);
				tmp1 = lamOut-lamIn;
				it++;
			}
			if(dk<=thr2){
				lamOut = D[i];
			}
		}else{
			lamIn = D[i+1] - rho*rho*0.1f;
			tmp2 = lamIn;
			gragg_d(D, U, rho, &tmp2, -1, n, 0, &fmid);
			lamOut = fminf(tmp2, D[i+1]-thr1);
			dkp1 = D[i+1]-lamOut;
			tmp1 = lamOut-lamIn;
			while((fabsf(tmp1)>thr1*dkp1 || it<1) && (dkp1>thr1) && (it<IT)){
				lamIn = lamOut;
				tmp2 = lamIn;
				gragg_d(D, U, rho, &tmp2, -1, n, 1, &tmp3);
				lamOut = fminf(tmp2, D[i+1]-thr1);
				dkp1 = fminf(dkp1,D[i+1]-lamOut);
				tmp1 = lamOut-lamIn;
				it++;
			}
			if(dkp1<=thr2){
				lamOut = D[i+1];
			}
		}
	}else{
		if(fabsf(D[i+1]-D[i])<=thr2){
			lamOut = D[i];
		}else{
			gragg_d(D, U, rho, &tmp2, i, n, 0, &fmid);
			lamIn = fmaxf(fminf(tmp2, D[i+1]-thr1), D[i]+thr1);
			tmp2 = lamIn;
			gragg_d(D, U, rho, &tmp2, i, n, 1, &tmp3);
			lamOut = fmaxf(fminf(tmp2, D[i+1]-thr1), D[i]+thr1);
			dkp1 = D[i+1]-lamOut;
			dk = lamOut-D[i];
			tmp1 = lamOut-lamIn;
			while((tmp1*(lamOut-lamIn)>0) && (it<IT)){
				tmp1 = lamOut-lamIn;
				lamIn = lamOut;
				tmp2 = lamIn;
				gragg_d(D, U, rho, &tmp2, i, n, 1, &tmp3);
				lamOut = fmaxf(fminf(tmp2, D[i+1]-thr1), D[i]+thr1);
				dkp1 = D[i+1]-lamOut;
				dk = lamOut-D[i];
				it++;
			}
			if(dk<=thr2){
				if(fabsf(fmid)>0.0f){
					if(fmid>0.0f){
						lamOut = D[i];
					}else{
						lamOut = D[i+1];
					}
				}else{
					lamOut = D[i];
				}
			}else{
				if(dkp1<=thr2){
					if(fabsf(fmid)>0.0f){
						if(fmid>0.0f){
							lamOut = D[i];
						}else{
							lamOut = D[i+1];
						}
					}else{              
						lamOut = D[i+1];
					}
				}
			}
		}
	}
	*lam = lamOut;
}


// The main routine 
__global__ void DiagSymdiag_k(float *R, float *d, float *s, int n){

	int flag, i, j, k, p, q, n2, nt, tj, Qj;
	int tidx = threadIdx.x%n;
	int Qt = (threadIdx.x-tidx)/n;
	int gb_index_x = Qt + blockIdx.x*(blockDim.x/n);
	float rho, lam, g, h;

	int steps = (int)log2f((float)n-1.0f);
	int IDX = (int)powf(2.0f,steps);
	k = n;

	n2 = n*n;
	nt = Qt*(2*n*(n+2)+2*IDX);

	extern __shared__ float sAds[];
	int* sI = (int*) &sAds[nt+2*n2+3*n];
	int* x0 = (int*) &sI[n];
	int* x1 = (int*) &x0[IDX];

	g = 0.00001f;
	h = 0.1f;

	for(i=0; i<n; i++){
		if(tidx==i){
			sAds[nt+i*n+tidx] =  d[gb_index_x*n+tidx];
		}else{
			if(tidx==i+1){
				sAds[nt+i*n+tidx] =  s[gb_index_x*n+tidx];
			}else{
				if(tidx==i-1){
					sAds[nt+i*n+tidx] =  s[gb_index_x*n+i];
				}else{
					sAds[nt+i*n+tidx] = 0.0f;
				}			
			}
		}
	}

	__syncthreads();

	////////////////////////// 2x2 stage
	if(tidx<IDX){
		x1[tidx]=n;
		j = IDX/2;
		x0[tidx]=0;

		for(i=0; i<steps; i++){
			IDX /= 2;
			if(tidx<j){
				x1[tidx] = x1[tidx]/2;
				j -= IDX/2;
				k /= 2;
			}else{
				x0[tidx] += x1[tidx]/2;
				x1[tidx] = k-x1[tidx]/2;
				j += IDX/2;
				k = k - k/2;
			}
		}
	}

	
	IDX = (int)powf(2.0f,steps);


	if((tidx%2)&&(tidx<IDX)){
		if (tidx<IDX-1){
			i = x0[tidx] + x1[tidx];
			sAds[nt+(i-1)*(n+1)] = sAds[nt+(i-1)*(n+1)] - sAds[nt+(i-1)*n+i];
			sAds[nt+i*(n+1)] = sAds[nt+i*(n+1)] - sAds[nt+(i-1)*n+i];
		}
	}

	__syncthreads();

	if((tidx%2)&&(tidx<IDX)){
		if(x1[tidx]>1){
			j = x0[tidx];
	            sAds[nt+(j-1)*(n+1)] = sAds[nt+(j-1)*(n+1)] - sAds[nt+(j-1)*n+j];
		      sAds[nt+j*(n+1)] = sAds[nt+j*(n+1)] - sAds[nt+(j-1)*n+j];
		}

		if(x1[tidx]==2){
	            j = x0[tidx];
			diagSym2D_d(sAds+nt+j*(n+1), sAds+nt+n2+j, n);
		}
	
		if(x1[tidx-1]==2){
		      j = x0[tidx-1];
			diagSym2D_d(sAds+nt+j*(n+1), sAds+nt+n2+j, n);
		}else{    
			if(x1[tidx]+x1[tidx-1]==2){
				j = x0[tidx-1];
				diagSym2D_d(sAds+nt+j*(n+1), sAds+nt+n2+j, n);
			}else{
				j = x0[tidx-1];
				sAds[nt+n2+j] = sAds[nt+j*(n+1)];
				sAds[nt+j*(n+1)] = 1.0f;
			}
		}
	}
	__syncthreads(); 

////////////////////////////////////////////////////////////
	for (q=2; q<=IDX; q*=2){

		// The division scheme
		flag = 0;
		k = n;
		j = n;
		Qj = 0;
		p = 0;
		tj = tidx;
		for(i=0; i<(int)log2f((float)IDX/q)+1; i++){
			k /= 2;
			if(tj>=k){
			  Qj += k;
			  tj -= k;
			  k = j-k;
			}
			if(i<(int)log2f((float)IDX/q)){
				j = k;
				p = Qj;
			}
		}
		if((j>=q) && (j>2)){
			flag = 1;
		}

		rho = sAds[nt+(p+j/2)*(n+1)-1];

		// Computes the u vector and creates its copy
		if(Qj==p){
			sAds[nt+n2+2*n+tidx] = sAds[nt+Qj*(n+1)+(k-1)*n+tj];
			sAds[nt+n2+3*n+tidx] = sAds[nt+Qj*(n+1)+(k-1)*n+tj];
		}else{
			sAds[nt+n2+2*n+tidx] = sAds[nt+Qj*(n+1)+tj];
			sAds[nt+n2+3*n+tidx] = sAds[nt+Qj*(n+1)+tj];	
		}
		
		// Initialization of the sorting indices array
		sI[tidx] = tidx-p;
		__syncthreads();

		// Once the value of rho is copied, puts 0 in its place in the matrix
		if(flag && (tj==0)){sAds[nt+(p+j/2-1)*(n+1)+1] = 0.0f;}
		if(flag && (tj==1)){sAds[nt+(p+j/2)*(n+1)-1] = 0.0f;}
		__syncthreads();

		// Performs the Jacobi rotation for identical eigenvalues
		if(flag){
			for(i=tidx+1; i<j+p; i++){
				jacoRot_d(sAds+nt+n2, sAds+nt+n2+3*n, sAds+nt,
						  rho, i, tidx, p, (p+j), n, g, h*fminf(32.0f/q,3.2f));
			}
		}
		__syncthreads();

		sAds[nt+n2+n+tidx] = sAds[nt+n2+tidx];

		// Sort the eigenvalues before solving the secular equation
		for(i=0; i<2*q; i++){
			__syncthreads();
			if(flag && (i%2) && (Qj<=p) && (tj<(j-1)/2)){
				eigSort_d(sAds+nt+n2, sAds+nt+n2+2*n, sI, p+2*tj+1);
			}else{
				if(flag && (Qj<=p) && (tj<j/2)){
					eigSort_d(sAds+nt+n2, sAds+nt+n2+2*n, sI, p+2*tj);
				}
			}
		}

		__syncthreads();

		sAds[nt+n2+4*n+tidx] = sAds[nt+n2+tidx];

		// Solving the secular equation
		if(flag){
			sAds[nt+n2+tidx] /= h;
		}
		__syncthreads();
		if(flag){
			if(rho>0){
				eigVal_d(sAds+nt+n2+p, sAds+nt+n2+2*n+p, &lam, sAds+nt+n2+3*n+p, 
					   sI+p, rho/h, tj+Qj-p, j, g);
			}else{
				eigVal_d(sAds+nt+n2+p, sAds+nt+n2+2*n+p, &lam, sAds+nt+n2+3*n+p, 
					   sI+p, rho/h, tj+Qj-p-1, j, g);
			}
		}
		__syncthreads();
		if(flag){
			sAds[nt+n2+tidx] = lam*h;
		}
		__syncthreads();

		// We should not have the same eigenvalue
		for(i=0; i<n; i++){
			if(tidx==0 && fabsf(sAds[nt+n2+i]-sAds[nt+n2+i+1])<h*thr2 && i<n-1){
				if(rho>0.0f){
					sAds[nt+n2+i+1] += 1.1f*h*thr2;
				}else{
					sAds[nt+n2+i] -= 1.1f*h*thr2;
				}
			}
		}

		__syncthreads();

		// Using Lowner Theorem
		if(flag){
			lowner_d(sAds+nt+n2+p, sAds+nt+n2+n+p, sAds+nt+n2+3*n+p,
					  sAds+nt+n2+2*n+p, sAds+nt+n2+4*n+p, rho, tj+Qj-p, 
					  j, g, h);
		}
		__syncthreads();

		// Deflation
		if(flag){
			deflation_d(sAds+nt+n2+p, sAds+nt+n2+4*n+p, sAds+nt+n2+2*n+p,
					  rho, sI+p, tj+Qj-p, j, h);
		}
		__syncthreads();

		// Compute eigenvectors
		if(flag){
			eigVect_d(sAds+nt+n2+p, sAds+nt+n2+n+p, sAds+nt+n2+2*n+p,
					  sAds+nt+n2+3*n+p*(n+1), rho, sI+p, tj+Qj-p, j, n, h);
		}
		__syncthreads();

		// Matrix multiplication
		for(i=0; i<2*q; i++){
			if(flag && i<j){
				prodLL_d(sAds+nt+n2+3*n+p*(n+1), sAds+nt+p*(n+1), i,
  					   tj+Qj-p, &lam, j, n);
			}
			__syncthreads();
			if(flag && i<j){
				sAds[nt+n2+3*n+p*(n+1)+i*n+tj+Qj-p] = lam;
			}
		}
		__syncthreads();
		
		// Matrix transposition
		if(flag){
			transLC_d(sAds+nt+n2+3*n+p*(n+1), sAds+nt+p*(n+1),
					  tj+Qj-p, j, n);
		}
		__syncthreads();
	}


	d[gb_index_x*n + tidx] = sAds[nt + n2 + tidx];
	for(i=0;i<n;i++){
			R[gb_index_x*n2+i*n+tidx] = sAds[nt+i*n+tidx];
	}
}