/**********************************************************************
Code associated to the paper:

Resolution of a large number of small random symmetric 
linear systems in single precision arithmetic on GPUs

by: 

Lokman A. Abbas-Turki and Stef Graillat

Those who re-use this code should mention in their code 
the name of the authors above.
**********************************************************************/


#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include "Hcyc.h"

/**********************************************************************

Only factorization

**********************************************************************/
///////////////////////////////////////////////////////////////////////
// Tridiagonalization: one thread per matrix
// This is a straight modification of function tred2 in 
// Numerical Recipes in C
///////////////////////////////////////////////////////////////////////
__global__ void tred_fact_k(float *a, float *d, float *s, int n)
{
	// The global memory access index
	int gb_index_x = threadIdx.x + blockIdx.x*blockDim.x;
	// Local integers
	int l, k, j, i, n2, nt;
	// Local floats
	float scale, hh, h, g, f;
	// Shared memory
	extern __shared__ float sAds[];

	n2 = n*n;
	nt = threadIdx.x*n*(n + 2);

	// Copy the matrix from global to shared memory
	for (i = 0; i < n; i++){
		for (j = 0; j < n; j++){
			sAds[nt + i*n + j] = a[gb_index_x*n*n + i*n + j];
		}
	}

	//Computing the tridiagonal form U.
	for (i = n - 1; i >= 1; i--) {
		l = i - 1;
		h = scale = 0.0f;
		if (l > 0) {
			for (k = 0; k <= l; k++)
				scale += fabs(sAds[nt + i*n + k]);
			if (scale < EPS) //Skip transformation.
				s[gb_index_x*n + i] = sAds[nt + i*n + l];
			else {
				for (k = 0; k <= l; k++) {
					sAds[nt + i*n + k] /= scale; //Use scaled As for transformation.
					h += sAds[nt + i*n + k] * sAds[nt + i*n + k]; //Compute sigma.
				}
				f = sAds[nt + i*n + l];
				g = (f >= 0.0f ? -sqrtf(h) : sqrtf(h));
				s[gb_index_x*n + i] = scale*g;
				h -= f*g; //Compute b.
				sAds[nt + i*n + l] = f - g; //Store u in the ith row of A.
				f = 0.0f;
				for (j = 0; j <= l; j++) {
					sAds[nt + j*n + i] = sAds[nt + i*n + j] / h; //Store u/b in the ith column of A.
					g = 0.0f; //Form an element of Au in g.
					for (k = 0; k <= j; k++)
						g += sAds[nt + j*n + k] * sAds[nt + i*n + k];
					for (k = j + 1; k <= l; k++)
						g += sAds[nt + k*n + j] * sAds[nt + i*n + k];
					sAds[nt + n2 + n + j] = g / h; //Form element of p in temporarily unused
					//element of e.
					f += sAds[nt + n2 + n + j] * sAds[nt + i*n + j];
				}
				hh = f / (h + h); //Compute B.
				for (j = 0; j <= l; j++) { //Form q and store it in e overwriting p.
					f = sAds[nt + i*n + j];
					sAds[nt + n2 + n + j] = g = sAds[nt + n2 + n + j] - hh*f;
					for (k = 0; k <= j; k++) //Reduce A.
						sAds[nt + j*n + k] -= (f*sAds[nt + n2 + n + k] + g*sAds[nt + i*n + k]);
				}
			}
		}
		else
			s[gb_index_x*n + i] = sAds[nt + i*n + l];
		sAds[nt + n2 + i] = h;
	}

	sAds[nt + n2] = 0.0f;
	s[gb_index_x*n] = 0.0f;

	//Accumulation of transformation matrices to get the product matrix Q.
	for (i = 0; i <= n - 1; i++) {
		l = i - 1;
		if (sAds[nt + n2 + i]) { //This block skipped when i=0.
			for (j = 0; j <= l; j++) {
				g = 0.0f;
				for (k = 0; k <= l; k++) //Use u and u/b stored in A to form HQ.
					g += sAds[nt + i*n + k] * sAds[nt + k*n + j];
				for (k = 0; k <= l; k++)
					sAds[nt + k*n + j] -= g*sAds[nt + k*n + i];
			}
		}
		d[gb_index_x*n + i] = sAds[nt + i*n + i];
		sAds[nt + i*n + i] = 1.0f; //Reset row and column of A to identity matrix for ...
		for (j = 0; j <= l; j++) sAds[nt + j*n + i] = sAds[nt + i*n + j] = 0.0f; // ... next iteration.
	}
	for (i = 0; i < n; i++){
		for (j = 0; j < n; j++){
			a[gb_index_x*n*n + i*n + j] = sAds[nt + i*n + j];
		}
	}
}


///////////////////////////////////////////////////////////////////////
// Tridiagonalization: The number of communicating threads per system 
// is equal to the size of the system. 
// This version performs much better than tred_k. 
///////////////////////////////////////////////////////////////////////
__global__ void tred_fact_max_k(float *a, float *d, float *s, int n)
{
	// Identifies the thread working within a group
	int tidx = threadIdx.x%n;
	// Identifies the data concerned by the computations
	int Qt = (threadIdx.x - tidx) / n;
	// The global memory access index
	int gb_index_x = Qt + blockIdx.x*(blockDim.x / n);
	// Local integers
	int i, j, k, l, n2, nt, nt1;
	// Local floats
	float scale, f, g, h;
	// Shared memory
	extern __shared__ float sAds[];

	n2 = n*n;
	nt = Qt*n*(n + 2);
	nt1 = (blockDim.x / n)*n*(n + 2);

	// Copy the matrix from global to shared memory
	for (i = 0; i < n; i++){
		sAds[nt + i*n + tidx] = a[gb_index_x*n2 + i*n + tidx];
	}
	__syncthreads();

	//Computing the tridiagonal form U.
	for (i = n - 1; i >= 1; i--) {
		l = i - 1;
		h = scale = 0.0f;
		if (tidx == 0){
			sAds[nt1 + Qt] = 0.0f;
		}
		if (l > 0) {
			for (k = 0; k <= l; k++)
				scale += fabs(sAds[nt + i*n + k]);
			if (scale < EPS){ //Skip transformation.
				if (tidx == 0){ s[gb_index_x*n + i] = sAds[nt + i*n + l]; }
			}
			else {
				if (tidx == 0){
					for (k = 0; k <= l; k++) {
						sAds[nt + i*n + k] /= scale; //Use scaled As for transformation.
						sAds[nt1 + Qt] += sAds[nt + i*n + k] * sAds[nt + i*n + k]; //Compute sigma.
					}
					f = sAds[nt + i*n + l];
					g = (f >= 0.0f ? -sqrtf(sAds[nt1 + Qt]) : sqrtf(sAds[nt1 + Qt]));
					s[gb_index_x*n + i] = scale*g;
					sAds[nt1 + Qt] -= f*g; //Compute b.
					sAds[nt + i*n + l] = f - g; //Store u in the ith row of A.
				}
				__syncthreads();
				if (tidx <= l){
					h = sAds[nt1 + Qt];
					sAds[nt + tidx*n + i] = sAds[nt + i*n + tidx] / h; //Store u/b in the ith column of A.
					g = 0.0f; //Form an element of Au in g.
					for (k = 0; k <= tidx; k++)
						g += sAds[nt + tidx*n + k] * sAds[nt + i*n + k];
					for (k = tidx + 1; k <= l; k++)
						g += sAds[nt + k*n + tidx] * sAds[nt + i*n + k];
					sAds[nt + n2 + n + tidx] = g / h; //Form element of p in temporarily unused
				}							//element of e.
				__syncthreads();
				if (tidx == 0){
					f = 0.0f;
					for (j = 0; j <= l; j++){
						f += sAds[nt + n2 + n + j] * sAds[nt + i*n + j];
					}
					sAds[nt1 + Qt] = f / (h + h); //Compute B.
				}
				__syncthreads();
				if (tidx <= l){ //Form q and store it in e overwriting p
					f = sAds[nt + i*n + tidx];
					g = sAds[nt + n2 + n + tidx] - sAds[nt1 + Qt] * f;
					sAds[nt + n2 + n + tidx] = g;
				}
				__syncthreads();

				if (tidx <= l){
					for (k = 0; k <= tidx; k++) //Reduce A.
						sAds[nt + tidx*n + k] -= (f*sAds[nt + n2 + n + k] + g*sAds[nt + i*n + k]);
				}
				__syncthreads();
			}
		}
		else{ if (tidx == 0) s[gb_index_x*n + i] = sAds[nt + i*n + l]; }
		if (tidx == 0) sAds[nt + n2 + i] = h;
		__syncthreads();
	}

	if (tidx == 0){
		sAds[nt + n2] = 0.0f;
		s[gb_index_x*n] = 0.0f;
	}
	__syncthreads();

	//Accumulation of transformation matrices to get the product matrix Q.
	for (i = 0; i <= n - 1; i++) {
		l = i - 1;
		if (sAds[nt + n2 + i]) { //This block skipped when i=0.
			if (tidx <= l){
				g = 0.0f;
				for (k = 0; k <= l; k++) //Use u and u/b stored in A to form HQ.
					g += sAds[nt + i*n + k] * sAds[nt + k*n + tidx];
				for (k = 0; k <= l; k++)
					sAds[nt + k*n + tidx] -= g*sAds[nt + k*n + i];
			}
			__syncthreads();
		}
		if (tidx == 0){
			d[gb_index_x*n + i] = sAds[nt + i*n + i];
			sAds[nt + i*n + i] = 1.0f; //Reset row and column of A to identity matrix for ...
		}

		if (tidx <= l){
			sAds[nt + tidx*n + i] = sAds[nt + i*n + tidx] = 0.0f; //... next iteration.
		}
		__syncthreads();
	}

	for (i = 0; i < n; i++){
		a[gb_index_x*n*n + i*n + tidx] = sAds[nt + i*n + tidx];
	}
}


///////////////////////////////////////////////////////////////////////
// Tridiagonalization: The number of communicating threads per system 
// is equal to ComT. This version performs better than tred_fact_k 
// but it is generally slower than tred_fact_max_k.
///////////////////////////////////////////////////////////////////////
__global__ void tred_fact_hyb_k(float *a, float *d, float *s, int n, int ComT)
{
	// Identifies the thread working within a group
	int tidx = threadIdx.x%ComT;
	// Identifies the data concerned by the computations
	int Qt = (threadIdx.x - tidx) / ComT;
	// The global memory access index
	int gb_index_x = Qt + blockIdx.x*(blockDim.x / ComT);
	// Local integers
	int i, j, k, l, n2, nt, nt1, p;
	// Local floats
	float scale, f, g, h;
	// Shared memory
	extern __shared__ float sAds[];

	n2 = n*n;
	nt = Qt*n*(n + 2);
	nt1 = (blockDim.x / ComT)*n*(n + 2);

	// Copy the matrix from global to shared memory
	for (i = 0; i < n; i++){
		for (p = 0; tidx + p < n; p += ComT){
			sAds[nt + i*n + tidx + p] = a[gb_index_x*n2 + i*n + tidx + p];
		}
	}
	__syncthreads();

	//Computing the tridiagonal form U.
	for (i = n - 1; i >= 1; i--) {
		l = i - 1;
		h = scale = 0.0f;
		if (tidx == 0){
			sAds[nt1 + Qt] = 0.0f;
		}
		if (l > 0) {
			for (k = 0; k <= l; k++)
				scale += fabs(sAds[nt + i*n + k]);
			if (scale < EPS){ //Skip transformation.
				if (tidx == 0){ s[gb_index_x*n + i] = sAds[nt + i*n + l]; }
			}
			else {
				if (tidx == 0){
					for (k = 0; k <= l; k++) {
						sAds[nt + i*n + k] /= scale; //Use scaled As for transformation.
						sAds[nt1 + Qt] += sAds[nt + i*n + k] * sAds[nt + i*n + k]; //Compute sigma.
					}
					f = sAds[nt + i*n + l];
					g = (f >= 0.0f ? -sqrtf(sAds[nt1 + Qt]) : sqrtf(sAds[nt1 + Qt]));
					s[gb_index_x*n + i] = scale*g;
					sAds[nt1 + Qt] -= f*g; //Compute b.
					sAds[nt + i*n + l] = f - g; //Store u in the ith row of A.
				}
				__syncthreads();
				for (p = 0; tidx + p < n; p += ComT){
					if (tidx + p <= l){
						h = sAds[nt1 + Qt];
						sAds[nt + (tidx + p)*n + i] = sAds[nt + i*n + tidx + p] / h; //Store u/b in the ith column of A.
						g = 0.0f; //Form an element of Au in g.
						for (k = 0; k <= tidx + p; k++)
							g += sAds[nt + (tidx + p)*n + k] * sAds[nt + i*n + k];
						for (k = tidx + p + 1; k <= l; k++)
							g += sAds[nt + k*n + tidx + p] * sAds[nt + i*n + k];
						sAds[nt + n2 + n + tidx + p] = g / h; //Form element of p in temporarily unused
					}							//element of e.
				}
				__syncthreads();
				if (tidx == 0){
					f = 0.0f;
					for (j = 0; j <= l; j++){
						f += sAds[nt + n2 + n + j] * sAds[nt + i*n + j];
					}
					sAds[nt1 + Qt] = f / (h + h); //Compute B.
				}
				__syncthreads();
				for (p = 0; tidx + p < n; p += ComT){
					if (tidx + p <= l){ //Form q and store it in e overwriting p
						g = sAds[nt + n2 + n + tidx + p] - sAds[nt1 + Qt] * sAds[nt + i*n + tidx + p];
						sAds[nt + n2 + n + tidx + p] = g;
					}
				}
				__syncthreads();
				for (p = 0; tidx + p < n; p += ComT){
					if (tidx + p <= l){
						for (k = 0; k <= tidx + p; k++) //Reduce A.
							sAds[nt + (tidx + p)*n + k] -= (sAds[nt + i*n + tidx + p] * sAds[nt + n2 + n + k] +
							sAds[nt + n2 + n + tidx + p] * sAds[nt + i*n + k]);
					}
				}
				__syncthreads();
			}
		}
		else{ if (tidx == 0) s[gb_index_x*n + i] = sAds[nt + i*n + l]; }
		if (tidx == 0) sAds[nt + n2 + i] = h;
		__syncthreads();
	}

	if (tidx == 0){
		sAds[nt + n2] = 0.0f;
		s[gb_index_x*n] = 0.0f;
	}
	__syncthreads();

	//Accumulation of transformation matrices to get the product matrix Q.
	for (i = 0; i <= n - 1; i++) {
		l = i - 1;
		if (sAds[nt + n2 + i]) { //This block skipped when i=0.
			for (p = 0; tidx + p < n; p += ComT){
				if (tidx + p <= l){
					g = 0.0f;
					for (k = 0; k <= l; k++) //Use u and u/b stored in A to form HQ.
						g += sAds[nt + i*n + k] * sAds[nt + k*n + tidx + p];
					for (k = 0; k <= l; k++)
						sAds[nt + k*n + tidx + p] -= g*sAds[nt + k*n + i];
				}
			}
			__syncthreads();
		}
		if (tidx == 0){
			d[gb_index_x*n + i] = sAds[nt + i*n + i];
			sAds[nt + i*n + i] = 1.0f; //Reset row and column of A to identity matrix for ...
		}

		for (p = 0; tidx + p < n; p += ComT){
			if (tidx + p <= l){
				sAds[nt + (tidx + p)*n + i] = sAds[nt + i*n + tidx + p] = 0.0f; //... next iteration.
			}
		}
		__syncthreads();
	}


	for (i = 0; i < n; i++){
		for (p = 0; tidx + p < n; p += ComT){
			a[gb_index_x*n*n + i*n + tidx + p] = sAds[nt + i*n + tidx + p];
		}
	}
}


/**********************************************************************

System resolution

**********************************************************************/
///////////////////////////////////////////////////////////////////////
// Parallel cyclic reduction tridiagonal symmetric matrices
// + two matrix/vector products
///////////////////////////////////////////////////////////////////////
__global__ void pcr_sym_k(float *q, float *a, float *b, float *y, int n)
{
	// Identifies the thread working within a group
	int tidx = threadIdx.x%n;
	// Identifies the data concerned by the computations
	int Qt = (threadIdx.x - tidx) / n;
	// The global memory access index
	int gb_index_x = Qt + blockIdx.x*(blockDim.x / n);
	// Local integers
	int i, n2, nt, lL, d, tL, tR;
	// Local floats
	float aL, bL, yL, sum, aLp, bLp, yLp;
	// Shared memory
	extern __shared__ float sAds[];

	n2 = n*n;
	nt = Qt*n*(n + 4);
	d = (n / 2 + (n % 2))*(tidx % 2) + (int)tidx / 2;

	float* sq = (float*)&sAds[nt];
	float* sa = (float*)&sq[n2];
	float* sb = (float*)&sa[n];
	float* sy = (float*)&sb[n];
	int* sl = (int*)&sy[n];

	sa[tidx] = a[gb_index_x*n + tidx];
	sb[tidx] = b[gb_index_x*n + tidx];
	sy[tidx] = y[gb_index_x*n + tidx];
	sl[tidx] = tidx;
	__syncthreads();

	//First matrix/vector product
	sum = 0.0f;
	for (i = 0; i < n; i++){
		sq[i*n + tidx] = q[gb_index_x*n2 + i*n + tidx];
		sum += sq[i*n + tidx] * sy[i];
	}
	__syncthreads();
	sy[tidx] = sum;

	//Left/Right indices of the reduction
	tL = tidx - 1;
	if (tL < 0) tL = 0;
	tR = tidx + 1;
	if (tR >= n) tR = 0;

	for (i = 0; i < (int)log2((float)n) + 1; i++){
		lL = (int)sl[tidx];

		aL = sa[tidx];
		bL = sb[tidx];
		yL = sy[tidx];

		bLp = sb[tL];

		//Reduction phase
		if (fabsf(aL) > EPS){
			aLp = sa[tL];
			yLp = sy[tL];

			//bL = b[tidx] - a[tidx]*c[tidx]/b[tidx-1];
			bL -= aL*aL / bLp;
			//yL = y[tidx] - a[tidx]*y[tidx-1]/b[tidx-1];
			yL -= aL*yLp / bLp;
			//aL = -a[tidx]*a[tidx-1]/b[tidx-1];
			aL = -aL*aLp / bLp;
		}

		aLp = sa[tR];
		bLp = sb[tR];
		if (fabsf(aLp) > EPS){
			yLp = sy[tR];
			//bL -= c[tidx+1]*a[tidx+1]/b[tidx+1];
			bL -= aLp*aLp / bLp;
			//yL -= c[tidx+1]*y[tidx+1]/b[tidx+1];
			yL -= aLp*yLp / bLp;
		}
		__syncthreads();

		//Permutation phase
		if (i < (int)log2((float)n)){
			sa[d] = aL;
			sb[d] = bL;
			sy[d] = yL;
			sl[d] = (int)lL;
			__syncthreads();
		}
	}

	sy[(int)tidx] = yL / bL;
	__syncthreads();

	sum = 0.0f;
	//Second matrix/vector product 
	//Non-coalescent access to recover the solution
	for (i = 0; i < n; i++){
		sum += sq[tidx*n + (int)sl[i]] * sy[i];
	}
	y[gb_index_x*n + tidx] = sum;
}