/**********************************************************************
Code associated to the paper:

Resolution of a large number of small random symmetric 
linear systems in single precision arithmetic on GPUs

by: 

Lokman A. Abbas-Turki and Stef Graillat

Those who re-use this code should mention in their code 
the name of the authors above.
**********************************************************************/

#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include "LDLt.h"
#include "Hcyc.h"
#include "rng.h"
#include "DC.h"

FILE *stream;


// Produces tridiagonal symmetric random matrices that satisfy 
// conditions 1) and 2) (if norm <= 10) precised at the top of the 
// file DC.cu
// The flag is used to decide whether the matrices should be 
// diagonally dominant or not.
__global__ void Tri_Rand_k(float *D, float *S, Tab2RNG_t *pt_RNG_In,
						   float norm, int i, int n, int L, bool flag)
{
	// The global access index for uniformly random values on (0,1)
	int idx = threadIdx.x + blockIdx.x*blockDim.x;
	// Identifies the thread working within a group
	int tidx = threadIdx.x%n;
	// Identifies the data concerned by the computations
	int Qt = (threadIdx.x - tidx) / n;
	// The global memory access index
	int gb_index_x = Qt + blockIdx.x*(blockDim.x / n);

	// Shared memory space
	extern __shared__ float R[];
	float *r0, *r1, R0, R1;
	r0 = R;
	r1 = R + blockDim.x;
	int tR = tidx + 1;
	if (tR > n-1) tR--;

	if(fabsf(pt_RNG_In[0][idx][0]-0.5f)<0.1f){
		R0 = 1.0f-pt_RNG_In[0][idx][0]/2.0f;
	}else{R0 = pt_RNG_In[0][idx][0];}

	if(fabsf(pt_RNG_In[0][idx][1]-0.5f)<0.1f){
		R1 = pt_RNG_In[0][idx][1]/2.0f;
	}else{R1 = pt_RNG_In[0][idx][1];}
	__syncthreads();


	// Communication needed in the case of positive matrices
	r0[threadIdx.x] = 2.0f*R0-1.0f;
	r1[threadIdx.x] = 2.0f*R1-1.0f;
	__syncthreads();

	if(gb_index_x*n + tidx + i < L){
		if (flag){
			R0 = fmaxf( fmaxf(fabsf(2.4f*r0[tidx + Qt*n]), 
						fabsf(2.3f*r1[tidx + Qt*n])), 
						fabsf(2.5f*r1[tR + Qt*n]) );
			R1 = fminf( fminf(fabsf(r0[tidx + Qt*n]), 
						fabsf(r1[tidx + Qt*n])),
						fabsf(r1[tR + Qt*n]) );
			R0 = norm*R0/2.5f;
			R1 = norm*copysignf(R1, r1[tidx + Qt*n])/2.5f;
			D[gb_index_x*n + tidx + i] = R0;
			if (tidx > 0){
				S[gb_index_x*n + tidx + i] = R1;
			}
		}
		else{
			D[gb_index_x*n + tidx + i] = norm*r0[threadIdx.x];
			if (tidx > 0){
				S[gb_index_x*n + tidx + i] = norm*r1[threadIdx.x];
			}
		}
	}
}


// Produces symmetric matrices from tridiagonal symmetric ones and 
// orthogonal matrices
__global__ void Sym_Trand_k(float *A, float *D, float *S, int n)
{
	// Identifies the thread working within a group
	int tidx = threadIdx.x%n;
	// Identifies the data concerned by the computations
	int Qt = (threadIdx.x - tidx) / n;
	// The global memory access index
	int gb_index_x = Qt + blockIdx.x*(blockDim.x / n);
	
	// Some local variables
	int i, j, n2, nt;
	n2 = n*n;
	nt = Qt*2*n2;

	// Shared memory space
	extern __shared__ float As[];
	float *A1, *A2, sum;
	A1 = As + nt;
	A2 = A1 + n2;

	for (i = 0; i < n; i++){
		A2[i*n + tidx] = A[gb_index_x*n2 + i*n + tidx];
		if (tidx==i+1){
			A1[i*n + tidx] = S[gb_index_x*n + tidx];
		}
		else{
			if (tidx == i - 1){
				A1[i*n + tidx] = S[gb_index_x*n + i];
			}
			else{
				A1[i*n + tidx] = 0.0f;
			}
		}
	}
	A1[tidx*n + tidx] = D[gb_index_x*n + tidx];
	__syncthreads();

	for (j = 0; j < n; j++){
		sum = 0.0f;
		for (i = 0; i < n; i++){
			sum += A2[tidx*n + i] * A1[j*n + i];
		}
		__syncthreads();
		A1[j*n + tidx] = sum;
	}
	__syncthreads();

	for (j = 0; j < n; j++){
		sum = 0.0f;
		for (i = 0; i < n; i++){
			sum += A2[tidx*n + i] * A1[i*n + j];
		}
		__syncthreads();
		A1[tidx*n + j] = sum;
	}
	__syncthreads();

	for (i = 0; i < n; i++){
		A[gb_index_x*n2 + i*n + tidx] = A1[i*n + tidx];
	}
}


int main(){

	int i, j, k;

	// The rank of the matrix
	int Dim = 32;
	// The minimum number of threads per block
	int minTB = 2;
	// The number of blocks
	int NB = 16384;
	// The number of matrices to invert
	int size = NB*minTB;

	// The matrix
	float *A, *AGPU;
	// The diagonal elements
	float *D, *DGPU;
	// The subdiagonal elements
	float *S, *SGPU;
	// The system vector
	float *Y, *YGPU;

	////////////////////////////////////////////////////////////
	// Memory allocation for RNG on CPU & GPU
	////////////////////////////////////////////////////////////
	cudaMalloc(&pt_CMRG, sizeof(TabSeedCMRG_t));
	cudaMalloc(&pt_2RNG, sizeof(Tab2RNG_t));
	pt_CMRGCPU = (TabSeedCMRG_t*) malloc(sizeof(TabSeedCMRG_t));

	PostInitDataCMRG();

	float TimerV;					// GPU timer instructions
	cudaEvent_t start, stop;			// GPU timer instructions
	cudaEventCreate(&start);			// GPU timer instructions
	cudaEventCreate(&stop);				// GPU timer instructions


	// Memory allocation
	A = (float *)calloc(size*Dim*Dim,sizeof(float));
	D = (float *)calloc(size*Dim,sizeof(float));
	S = (float *)calloc(size*Dim,sizeof(float));
	Y = (float *)calloc(size*Dim,sizeof(float));
	cudaMalloc(&AGPU, size*Dim*Dim*sizeof(float));
	cudaMalloc(&DGPU, size*Dim*sizeof(float));
	cudaMalloc(&SGPU, size*Dim*sizeof(float));
	cudaMalloc(&YGPU, size*Dim*sizeof(float));


	int IDX = 2*((int)pow(2.0,(int)log2((double)Dim-1.0)));
	int HM = M/(NB*Dim);
	for (i=0; i<minTB+HM; i+=HM){
		UniformRandom();
		Tri_Rand_k <<<NB,HM*Dim,2*HM*Dim*sizeof(float)>>>(DGPU, SGPU, pt_2RNG, 
										  10.0f, i*Dim*NB, Dim, 
										  Dim*minTB*NB, true);
	}

	cudaMemcpy(D, DGPU, size*Dim*sizeof(float), cudaMemcpyDeviceToHost);
	cudaMemcpy(S, SGPU, size*Dim*sizeof(float), cudaMemcpyDeviceToHost);

	/////////////////////////// LDLt on Random matrices ////////////////////////////
	////////////////// To uncomment if it has to be implemented ////////////////////
	for (i=0; i<size; i++){
		for (j=0; j<Dim; j++){
			for (k=0; k<Dim; k++){
				if(j==k){
				A[i*Dim*Dim+j*Dim+k] = 1.0f;
				}else{A[i*Dim*Dim+j*Dim+k] = 0.5f;}
			}
		}
	}
	for (i=0; i<size; i++){
		for (j=0; j<Dim; j++){
			Y[j+i*Dim]=0.5f*j;
		}
	}
	cudaMemcpy(AGPU,A,size*Dim*Dim*sizeof(float),cudaMemcpyHostToDevice);
	tred_fact_max_k<<<NB,Dim*minTB,(size*Dim*(Dim+2)/NB+minTB)*sizeof(float)>>>(AGPU, DGPU, 
													    SGPU, Dim);
	cudaMemcpy(YGPU,Y,size*Dim*sizeof(float),cudaMemcpyHostToDevice);
	cudaMemcpy(DGPU,D,size*Dim*sizeof(float),cudaMemcpyHostToDevice);
	cudaMemcpy(SGPU,S,size*Dim*sizeof(float),cudaMemcpyHostToDevice);

	Sym_Trand_k <<<NB, Dim*minTB, 2 * Dim*Dim*minTB*sizeof(float) >>>(AGPU, DGPU, SGPU, Dim);
	cudaMemcpy(A, AGPU, size*Dim*Dim*sizeof(float), cudaMemcpyDeviceToHost);
	fopen_s(&stream, "val.txt", "w");
	for (i = 0; i<size; i++){
		if (i == 5273){ // One can check with other matrices
			for (j = 0; j<Dim; j++){
				for (k = 0; k<Dim; k++){
					fprintf(stream, "%f,", A[i*Dim*Dim + j*Dim + k]);
				}
				fprintf(stream, "\n\n");
			}
			fprintf(stream, "\n");
		}
	}
	fclose(stream);

	// The number of communicating threads per block in the hybrid solution
	int ComT = Dim/2;
	cudaEventRecord(start,0);
	LDLt_hyb_k<<<NB,ComT*minTB,minTB*((Dim*Dim+Dim)/2+Dim)*sizeof(float)>>>(AGPU, YGPU, 
													Dim, ComT);
	
	// The other solution
	//LDLt_max_k<<<NB,Dim*minTB,minTB*((Dim*Dim+Dim)/2+Dim)*sizeof(float)>>>(AGPU, YGPU, Dim);
	//LDLt_k<<<NB,minTB,minTB*((Dim*Dim+Dim)/2+Dim)*sizeof(float)>>>(AGPU, YGPU, Dim);
	cudaEventRecord(stop,0);
	cudaEventSynchronize(stop);
	cudaEventElapsedTime(&TimerV,start, stop);
	cudaMemcpy(Y, YGPU, size*Dim*sizeof(float), cudaMemcpyDeviceToHost);
	for (i=0; i<size; i++){
		if(i==5273){
			printf("\n\n");
			for (j=0; j<Dim; j++){
				printf("%f \n",Y[j+i*Dim]);
			}
		}
	}
	printf("The symmetric matrix is available in the file val.txt \n");
	printf("Execution time of the LDLt: %f ms\n", TimerV);


	/////////// Householder tridiagonalization + PCR on Random matrices ////////////
	////////////////// To uncomment if it has to be implemented ////////////////////
	/*******************************************************************************
	for (i=0; i<size; i++){
		for (j=0; j<Dim; j++){
			for (k=0; k<Dim; k++){
				if(j==k){
				A[i*Dim*Dim+j*Dim+k] = 1.0f;
				}else{A[i*Dim*Dim+j*Dim+k] = 0.5f;}
			}
		}
	}
	for (i=0; i<size; i++){
		for (j=0; j<Dim; j++){
			Y[j+i*Dim]=0.5f*j;
		}
	}
	cudaMemcpy(AGPU,A,size*Dim*Dim*sizeof(float),cudaMemcpyHostToDevice);
	tred_fact_max_k<<<NB,Dim*minTB,(size*Dim*(Dim+2)/NB+minTB)*sizeof(float)>>>(AGPU, DGPU, 
													    SGPU, Dim);
	cudaMemcpy(YGPU,Y,size*Dim*sizeof(float),cudaMemcpyHostToDevice);
	cudaMemcpy(DGPU,D,size*Dim*sizeof(float),cudaMemcpyHostToDevice);
	cudaMemcpy(SGPU,S,size*Dim*sizeof(float),cudaMemcpyHostToDevice);

	Sym_Trand_k <<<NB, Dim*minTB, 2 * Dim*Dim*minTB*sizeof(float) >>>(AGPU, DGPU, SGPU, Dim);
	cudaMemcpy(A, AGPU, size*Dim*Dim*sizeof(float), cudaMemcpyDeviceToHost);
	fopen_s(&stream, "val.txt", "w");
	for (i = 0; i<size; i++){
		if (i == 5273){ // One can check with other matrices
			for (j = 0; j<Dim; j++){
				for (k = 0; k<Dim; k++){
					fprintf(stream, "%f,", A[i*Dim*Dim + j*Dim + k]);
				}
				fprintf(stream, "\n\n");
			}
			fprintf(stream, "\n");
		}
	}
	fclose(stream);

	cudaEventRecord(start,0);
	tred_fact_max_k<<<NB,Dim*minTB,(size*Dim*(Dim+2)/NB+minTB)*sizeof(float)>>>(AGPU, DGPU, 
													    SGPU, Dim);
	// The SIMD solution
	//tred_fact_k<<<NB,size/NB,size*Dim*(Dim+2)*sizeof(float)/NB>>>(AGPU, DGPU, SGPU, Dim);

	pcr_sym_k<<<NB,Dim*minTB,(minTB*Dim*(Dim+4))*sizeof(float)>>>(AGPU, SGPU, DGPU, YGPU, Dim);
	cudaEventRecord(stop,0);
	cudaEventSynchronize(stop);
	cudaEventElapsedTime(&TimerV,start, stop);

	cudaMemcpy(Y, YGPU, size*Dim*sizeof(float), cudaMemcpyDeviceToHost);
	for (i=0; i<size; i++){
		if(i==5273){
			printf("\n\n");
			for (j=0; j<Dim; j++){
				printf("%f \n",Y[j+i*Dim]);
			}
		}
	}
	printf("The symmetric matrix is available in the file val.txt \n");
	printf("Execution time of the Householder + PCR: %f ms\n", TimerV);
	*******************************************************************************/


	//////////////////// Divide and Conquer on Random matrices /////////////////////
	////////////////// To uncomment if it has to be implemented ////////////////////
	/*******************************************************************************
	fopen_s( &stream, "tri.txt", "w" );
	for (i = 0; i<size; i++){
		if (i == 5273){ // One can check with other matrices
			for (j = 0; j<Dim; j++){
				for (k = 0; k<Dim; k++){
					if (j == k){
						fprintf( stream, "%f,", D[j+i*Dim]);
					}else{
						if (j == k + 1){
							fprintf( stream, "%f,", S[j+i*Dim]);
						}else{ 
							if (j + 1 == k ){
								fprintf( stream, "%f,", S[k+i*Dim]);
							}else{ 
								fprintf( stream, "%f,", 0.0f);
							}
						}
					}
				}
				fprintf( stream, "\n");
			}
		}
	}
	printf("\n");
	fclose( stream );
	
	cudaEventRecord(start,0);
	DiagSymdiag_k<<<NB,Dim*minTB,(minTB*(2*Dim*(Dim+2)+IDX))*sizeof(float)>>>(AGPU, DGPU, 
													  SGPU, Dim);
	cudaEventRecord(stop,0);
	cudaEventSynchronize(stop);
	cudaEventElapsedTime(&TimerV,start, stop);

	cudaMemcpy(D, DGPU, size*Dim*sizeof(float), cudaMemcpyDeviceToHost);
	cudaMemcpy(A, AGPU, size*Dim*Dim*sizeof(float), cudaMemcpyDeviceToHost);
	fopen_s(&stream, "val.txt", "w");
	for (i = 0; i<size; i++){
		if (i == 5273){ // One can check with other matrices
			for (j = 0; j<Dim; j++){
				for (k = 0; k<Dim; k++){
					fprintf(stream, "%f,", A[i*Dim*Dim + j*Dim + k]);
				}
				fprintf(stream, "\n\n");
			}
			fprintf(stream, "\n");
		}
	}
	fclose(stream);
	for (i = 0; i<size; i++){
		if (i == 5273){ // One can check with other matrices
			for (j = 0; j<Dim; j++){
				printf("Eigenvalue %i is: %f \n", j, D[j + i*Dim]);
			}
		}
	}
	printf("The tridiagonal matrix is available in the file tri.txt \n");
	printf("The eigenvectors are available in the file val.txt \n");
	printf("Execution time of the D&C: %f ms\n", TimerV);
	*******************************************************************************/
	

	// Memory free for RNG on CPU & GPU
	cudaFree(pt_CMRG);
	cudaFree(pt_2RNG);
	free(pt_CMRGCPU);

	// Memory free for other arrays
	free(A);
	cudaFree(AGPU);
	free(D);
	cudaFree(DGPU);
	free(S);
	cudaFree(SGPU);
	free(Y);
	cudaFree(YGPU);

	cudaEventDestroy(start);			// GPU timer instructions
	cudaEventDestroy(stop);				// GPU timer instructions

	return 0;
}