/**********************************************************************
Code associated to the paper:

Resolution of a large number of small random symmetric 
linear systems in single precision arithmetic on GPUs

by: 

Lokman A. Abbas-Turki and Stef Graillat

Those who re-use this code should mention in their code 
the name of the authors above.
**********************************************************************/

#include "rng.h"

////////////////////////////////////////////////////////////////
// Memory for RNG use 
////////////////////////////////////////////////////////////////
// The state variables of CMRG on GPU 
TabSeedCMRG_t *pt_CMRG;
// The combination of CMRGs, the used RNGs
Tab2RNG_t *pt_2RNG;
// The state variables of CMRG on CPU
TabSeedCMRG_t *pt_CMRGCPU;
// Matrixes associated to the post treatment of the CMRG
// - First MRG
double A1[3][3];
// - Second MRG
double A2[3][3];

////////////////////////////////////////////////////////////////
// This kernel combines the two MRGs
////////////////////////////////////////////////////////////////
__global__ void Comb_CMRG_k(int comb_idx,
							TabSeedCMRG_t *pt_CMRG_In, 
						    Tab2RNG_t *pt_Comb_Out)
{

 // The global access index
 int gb_index_x = threadIdx.x + blockIdx.x*blockDim.x;

 const int m1 = 2147483647; 
 int inter;               

 // - allocate intermediate tables
 int CMRG2;
 int CMRG5;
 // - copy lines of data from global to registers
 CMRG2 = pt_CMRG_In[0][gb_index_x][2];
 CMRG5 = pt_CMRG_In[0][gb_index_x][5];

 if (CMRG2 < CMRG5) {
   inter = CMRG2 - CMRG5 + m1;
 } else {
   inter = CMRG2 - CMRG5;
 }
 if (inter == 0){
   pt_Comb_Out[0][gb_index_x][comb_idx] = Invmp*m1;
 } else {
   pt_Comb_Out[0][gb_index_x][comb_idx] = Invmp*inter;
 }
}

////////////////////////////////////////////////////////////////
// This kernel generates the state values of two MRGs
////////////////////////////////////////////////////////////////
__global__ void Rand_CMRG_k(TabSeedCMRG_t *pt_CMRG_InOut)
{
	
  // The global access index
 int gb_index_x = threadIdx.x + blockIdx.x*blockDim.x;

 const int m1 = 2147483647;// Requested for the simulation
 const int m2 = 2145483479;// Requested for the simulation
 int h, p12, p13, p21, p23;// Requested local parameters

 // First Component 
 h = pt_CMRG_InOut[0][gb_index_x][0]/q13; 
 p13 = a13*(h*q13-pt_CMRG_InOut[0][gb_index_x][0])-h*r13;

 h = pt_CMRG_InOut[0][gb_index_x][1]/q12; 
 p12 = a12*(pt_CMRG_InOut[0][gb_index_x][1]-h*q12)-h*r12;

 if (p13 < 0) {
   p13 = p13 + m1;
 }
 if (p12 < 0) {
   p12 = p12 + m1;
 }
  
 pt_CMRG_InOut[0][gb_index_x][0] = pt_CMRG_InOut[0][gb_index_x][1];
 pt_CMRG_InOut[0][gb_index_x][1] = pt_CMRG_InOut[0][gb_index_x][2];

 if( (p12 - p13) < 0){
   pt_CMRG_InOut[0][gb_index_x][2] = p12 - p13 + m1;  
 } else {
   pt_CMRG_InOut[0][gb_index_x][2] = p12 - p13;
 }
  
 // Second Component 
 h = pt_CMRG_InOut[0][gb_index_x][3]/q23; 
 p23 = a23*(h*q23-pt_CMRG_InOut[0][gb_index_x][3])-h*r23;

 h = pt_CMRG_InOut[0][gb_index_x][5]/q21; 
 p21 = a21*(pt_CMRG_InOut[0][gb_index_x][5]-h*q21)-h*r21;

 if (p23 < 0){
   p23 = p23 + m2;
 }
 if (p12 < 0){
   p21 = p21 + m2;
 }
  
 pt_CMRG_InOut[0][gb_index_x][3] = pt_CMRG_InOut[0][gb_index_x][4];
 pt_CMRG_InOut[0][gb_index_x][4] = pt_CMRG_InOut[0][gb_index_x][5];

 if ( (p21 - p23) < 0) {
   pt_CMRG_InOut[0][gb_index_x][5] = p21 - p23 + m2;  
 } else {
   pt_CMRG_InOut[0][gb_index_x][5] = p21 - p23;
 }  
}

////////////////////////////////////////////////////////////////
// Generation of an array of two uniformly distributed variables
////////////////////////////////////////////////////////////////
void UniformRandom(void)
{
 // GPU thread management variables
 // - Grid and block features
 dim3 Dg(M/BlockSize,1,1); 
 dim3 Db(BlockSize,1,1);
  
 // Call the kernel and execute it on the GPU 
 // Getting two arrays (0,1) of uniformly distributed random variables 

 cudaFuncSetCacheConfig(Rand_CMRG_k,cudaFuncCachePreferL1);
 Rand_CMRG_k<<<Dg,Db>>>(pt_CMRG);
 cudaFuncSetCacheConfig(Comb_CMRG_k,cudaFuncCachePreferL1);
 Comb_CMRG_k<<<Dg,Db>>>(0, pt_CMRG, pt_2RNG);
 cudaFuncSetCacheConfig(Rand_CMRG_k,cudaFuncCachePreferL1);
 Rand_CMRG_k<<<Dg,Db>>>(pt_CMRG);
 cudaFuncSetCacheConfig(Comb_CMRG_k,cudaFuncCachePreferL1);
 Comb_CMRG_k<<<Dg,Db>>>(1, pt_CMRG, pt_2RNG);
}

// Functions from http://www.iro.umontreal.ca/~lecuyer/myftp/streams00/c/

/*-------------------------------------------------------------------------*/

static double MultModM (double a, double s, double c, double m)
   /* Compute (a*s + c) % m. m must be < 2^35.  Works also for s, c < 0 */
{
   double v;
   long a1;
   v = a * s + c;
   if ((v >= two53) || (v <= -two53)) {
      a1 = (long) (a / two17);
      a -= a1 * two17;
      v = a1 * s;
      a1 = (long) (v / m);
      v -= a1 * m;
      v = v * two17 + a * s + c;
   }
   a1 = (long) (v / m);
   if ((v -= a1 * m) < 0.0)
      return v += m;
   else
      return v;
}


/*-------------------------------------------------------------------------*/

static void MatVecModM (double A[3][3], double s[3], double v[3], double m)
   /* Returns v = A*s % m.  Assumes that -m < s[i] < m. */
   /* Works even if v = s. */
{
   int i;
   double x[3];
   for (i = 0; i < 3; ++i) {
      x[i] = MultModM (A[i][0], s[0], 0.0, m);
      x[i] = MultModM (A[i][1], s[1], x[i], m);
      x[i] = MultModM (A[i][2], s[2], x[i], m);
   }
   for (i = 0; i < 3; ++i)
      v[i] = x[i];
}


/*-------------------------------------------------------------------------*/

static void MatMatModM (double A[3][3], double B[3][3], double C[3][3],
                        double m)
   /* Returns C = A*B % m. Work even if A = C or B = C or A = B = C. */
{
   int i, j;
   double V[3], W[3][3];
   for (i = 0; i < 3; ++i) {
      for (j = 0; j < 3; ++j)
         V[j] = B[j][i];
      MatVecModM (A, V, V, m);
      for (j = 0; j < 3; ++j)
         W[j][i] = V[j];
   }
   for (i = 0; i < 3; ++i) {
      for (j = 0; j < 3; ++j)
         C[i][j] = W[i][j];
   }
}


/*-------------------------------------------------------------------------*/

static void MatPowModM (double A[3][3], double B[3][3], double m, long n)
   /* Compute matrix B = A^n % m ;  works even if A = B */
{
   int i, j;
   double W[3][3];

   /* initialize: W = A; B = I */
   for (i = 0; i < 3; i++) {
      for (j = 0; j < 3; ++j) {
         W[i][j] = A[i][j];
         B[i][j] = 0.0;
      }
   }
   for (j = 0; j < 3; ++j)
      B[j][j] = 1.0;

   /* Compute B = A^n % m using the binary decomposition of n */
   while (n > 0) {
      if (n % 2)
         MatMatModM (W, B, B, m);
      MatMatModM (W, W, W, m);
      n /= 2;
   }
}

////////////////////////////////////////////////////////////////
// Post initialization of CMRG
////////////////////////////////////////////////////////////////
void PostInitDataCMRG()
{
 const int m1 = 2147483647; // Requested for the simulation
 const int m2 = 2145483479; // Requested  for the simulation
 int j;					// loop indices

 // + Init of the posttreatment table for CMRG rng (should
 //   be turned into a file reading in the future)
   A1[0][0] = 2119409629.0; A1[0][1] =  302707381.0; A1[0][2] =  655487731.0;
   A1[1][0] =  946145520.0; A1[1][1] = 1762689149.0; A1[1][2] =  302707381.0;
   A1[2][0] = 1139076568.0; A1[2][1] =  600956040.0; A1[2][2] = 1762689149.0;

   A2[0][0] = 1705536521.0; A2[0][1] = 1409357255.0; A2[0][2] = 1489714515.0;
   A2[1][0] = 1443451163.0; A2[1][1] = 1705536521.0; A2[1][2] = 1556328147.0;
   A2[2][0] = 1624922073.0; A2[2][1] = 1443451163.0; A2[2][2] =  130172503.0;

 MatPowModM (A1, A1, m1, 33554432);
 MatPowModM (A2, A2, m2, 33554432);
 
////////////////////////////////////////////////////////////////
// Local variables used in the initialization 
////////////////////////////////////////////////////////////////
 double s1[3] = {1.0, 1.0, 1.0};
 double s2[3] = {1.0, 1.0, 1.0};

 for (j = 0; j < M; j++) {
	   pt_CMRGCPU[0][j][0] = (int) s1[0]; 
	pt_CMRGCPU[0][j][1] = (int) s1[1];
	   pt_CMRGCPU[0][j][2] = (int) s1[2];
	   pt_CMRGCPU[0][j][3] = (int) s2[0];
	   pt_CMRGCPU[0][j][4] = (int) s2[1];
	   pt_CMRGCPU[0][j][5] = (int) s2[2];
	   MatVecModM (A1, s1, s1, m1);
       MatVecModM (A2, s2, s2, m2);
 }
 
 // - Copy CMRG data on the GPU
 cudaMemcpy(pt_CMRG, pt_CMRGCPU, sizeof(TabSeedCMRG_t), cudaMemcpyHostToDevice);
}
